/*
 *   (c) 2003 by Eduardo Enrique Gonzlez Rodrguez
 *   Distributed under GPL, see license.txt file in the package
 *
 *   Entropy coding by redundancy feedback; implementation of the
 * dynamic rf decoding algorithms.
 */

#include "defs.h"
#include "dec_dyn.h"
#include "freqs.h"
#include <stdlib.h>
#include <string.h>

extern int bit_mask[];
extern unsigned char bits_for_range[BR_TABLE_SIZE];

#include "in.inc" /* Bit and number reading functions */
#include "out.inc" /* write_number_ac() */

/*
 *   Creates a virtual symbol block out of a frequency table; it will be
 * used in the decoding process, since all blocks require some information
 * from a prior one in order to process their data. It will not be written
 * to the output though, it will be referred to while decoding the first
 * symbol block.
 */

void
dec_dyn_block0(word *freq, symbol *block, int block_size, word *counter,
	       word *last_counter)
{
	int i;
	unsigned char j = -1;

	memset(last_counter, 0, sizeof(last_counter[0]) * MAX_SYMBOLS);
	scale_freq(freq, block_size);
	for (i = 0; i < SLIDINGW_SIZE; i++) {
		do
			j++;
		while (!freq[j]);
		block[i] = j;
		counter[i] = last_counter[j]++;
		freq[j]--;
	}
	for (i = 0; i < MAX_SYMBOLS; i++) {
		last_counter[i] -= (last_counter[i]? 1: 0);
		freq[i] = last_counter[i];
		freq[i]++;
	}
	return;
}

/*
 *   Decodes a data block, using a dynamic estimation of the frequency
 * distribution of all symbols and phasing in codes in the feedback stage.
 * This function can be used to operate on a data stream on block mode.
 *
 *   Steps taken are:
 *
 *   1.- Loads block length
 *   2.- Until the whole sliding window has been rewritten:
 *
 *         2.a)  Load codeword (pointer to the sliding window)
 *         2.b)  In case (codeword == output pointer) {
 *                   **** A rare symbol has been found ****
 *                   - Read a symbol from the input stream
 *                   - Put that symbol in the sliding window
 *                   - Update the model, introducing the symbol in it
 *               }
 *         2.c)  Else {
 *                   **** A common symbol has been found ****
 *                   - Put in the sliding window the symbol pointed to by
 *                   the codeword
 *                   - Get the the frequency of the symbol and the
 *                   ordinal of that symbol occurrence in the sliding window
 *                   - Feedback the ordinal as a phasing in code
 *                   - Update the model, introducing the symbol in it
 *               }
 *
 *   Model update basically means to change the frequency tables accordingly
 * (symbols entering the sliding window increased, symbols leaving decreased).
 * Also the address and last_number tables must be modified, so for every
 * index to the sliding window, we can determine the order of the referred
 * symbol.
 */

void
dec_dyn_pic(stream *input, dword *input_idx, word *counter,
	    word *last_counter, word *freq, unsigned char *output,
	    dword *output_size)
{
	dword i;
	dword codeword;
	word sym;
	int k;
	word end; /* Number of syms at input */
	dword index;

	index = *input_idx;
	end = read_number_pic(input, &index, SLIDINGW_SIZE) + 1;
	codeword = read_bits(input, &index, SLIDINGW_BITS);
	for (i = 1; i != end + 1; i++) {
		codeword = (codeword - 21065) & ((1 << SLIDINGW_BITS) - 1);
		if (codeword == 0) {
			sym = read_bits(input, &index, SYMBOL_BITS);
			freq[(int)output[i]]--; /* Update freqs */
			output[i] = sym;
			freq[sym] = 1;
			last_counter[sym] = 0;
			counter[i] = 0;
			codeword = read_bits(input, &index, SLIDINGW_BITS);
			continue;
		}
		sym = output[(int)codeword];
		/* Feedback redundant information of codeword */
		codeword = (word)(last_counter[(int)sym] - counter[codeword]);
		k = FRLOG2(freq[(int)sym]);
		k += ((codeword | (1 << k)) < freq[(int)sym])? 1: 0;
		codeword |= (read_bits(input, &index, SLIDINGW_BITS - k)
			     << k); /* Read some bits to get next codeword */
		freq[(int)output[i]]--; /* Updates freqs */
		freq[(int)sym]++;
		last_counter[(int)sym]++; /* Updates rest of model */
		counter[i] = last_counter[(int)sym];
		output[i] = sym; /* Writes decoded symbol */
	}
	*input_idx = index;
	*output_size = i - 1;
	return;
}

/*
 *   Modification of the previous function to feedback information using
 * arithmetic coding instead of phasing in codes.
 */

void
dec_dyn_ac(stream *input, dword *input_idx, word *counter,
	   word *last_counter, word *freq, unsigned char *output,
	   dword *output_idx)
{
	dword i;
	dword codeword;
	word sym;
	word end; /* Number of syms at input */
	dword index;

	index = *input_idx;
	end = read_number_pic(input, &index, SLIDINGW_SIZE) + 1;
	codeword = read_bits(input, &index, SLIDINGW_BITS);
	for (i = 1; i != end + 1; i++) {
		if (codeword == 0) {
			sym = read_bits(input, &index, SYMBOL_BITS);
			freq[(int)output[i]]--; /* Update freqs */
			output[i] = sym;
			freq[sym] = 1;
			last_counter[sym] = 0;
			counter[i] = 0;
			codeword = read_bits(input, &index, SLIDINGW_BITS);
			continue;
		}
		sym = output[(int)codeword];
		/* Feedback redundant information of codeword */
		codeword = (word)(last_counter[(int)sym] - counter[codeword]);
		write_number_ac(input, &index, codeword, freq[sym]);
		codeword = read_bits(input, &index, SLIDINGW_BITS);
		freq[(int)output[i]]--; /* Updates freqs */
		freq[(int)sym]++;
		last_counter[(int)sym]++; /* Updates rest of model */
		counter[i] = last_counter[(int)sym];
		output[i] = sym; /* Writes decoded symbol */
	}
	*input_idx = index;
	*output_idx = i - 1;
	return;
}

/*
 *   Modification of the previous function to work without any previous
 * block processed before. It has to be called to decode the very first
 * block of symbols of any input stream.
 */

void
dec_dyn1st(stream *input, dword *input_idx, word *counter,
	   word *last_counter, word *freq, unsigned char *output,
	   dword *output_idx)
{
	dword i;
	dword codeword;
	word sym;
	word end; /* Number of syms at input */
	dword index;

	index = *input_idx;
	end = read_number_pic(input, &index, SLIDINGW_SIZE) + 1;
	/* Reads first symbol directly */
	sym = read_bits(input, &index, SYMBOL_BITS);
	output[1] = sym;
	last_counter[output[1]] = 0;
	freq[output[1]] = 1;
	counter[1] = 0;

	codeword = read_number_ac(input, &index, 2);
	for (i = 2; i != end + 1; i++) {
		if (codeword == 0) {
			sym = read_bits(input, &index, SYMBOL_BITS);
			output[i] = sym;
			freq[sym] = 1;
			last_counter[sym] = 0;
			counter[i] = 0;
			codeword = read_number_ac(input, &index, i + 1);
			continue;
		}
		sym = output[(int)codeword];
		/* Feedback redundant information of codeword */
		codeword = (word)(last_counter[(int)sym] - counter[codeword]);
		write_number_ac(input, &index, codeword, freq[sym]);
		codeword = read_number_ac(input, &index, i + 1);
		freq[(int)sym]++;
		last_counter[(int)sym]++; /* Updates rest of model */
		counter[i] = last_counter[(int)sym];
		output[i] = sym; /* Writes decoded symbol */
	}
	*input_idx = index;
	*output_idx = i - 1;
	return;
}


