/*
 *   (c) 2003 by Eduardo Enrique Gonzlez Rodrguez
 *   Distributed under GPL, see license.txt file in the package
 *
 *   Entropy coding by redundancy feedback; implementation of the
 * static frequency estimation decoding algorithms.
 */

#include "defs.h"
#include "dec_stat.h"
#include "freqs.h"
#include <stdlib.h>
#include <string.h>

extern int bit_mask[];
extern unsigned char bits_for_range[BR_TABLE_SIZE];

#include "in.inc" /* Bit and number reading functions */
#include "out.inc" /* write_number_ac() */


/*
 *   Creates a decode table out of gathered and adjusted frequencies, with
 * a specified table size. The decode table is simply an array where every
 * record contains three entries bitpacked.
 *
 *   Each entry has the following information:
 *
 *    +-----------------------+--------------+--------+
 *    | Redundant information | Bits to read | Symbol |
 *    +-----------------------+--------------+--------+
 *
 *   The most significant field keeps the redundant information associated
 * to the codeword. The other fields keep the number of bits to read in
 * order to complete the next codeword and the symbol to output.
 */

static void
build_dec_table_pic(word *freq, dword *dec_table, int table_bits, word *hash)
{
	int i;
	int sym;
	int nbits;
	int k;
	int c;
	int n;
	int j = 0;

	/* Fill primary table */
	for (sym = 0; sym != MAX_SYMBOLS; sym++) {
		nbits = FRLOG2(freq[sym]);
		c = (1 << (nbits + 1)) - freq[sym];
		for (i = 0; i < freq[sym]; i++) {
			if (i < c) {
				n = i;
				k = nbits;
			} else {
				n = ((i + c) & -2) | ((i - c) & 1);
				k = nbits + 1;
			}
			n <<= (table_bits - k + NBITS_BITS + SYMBOL_BITS);
			dec_table[hash[j]] = ((table_bits - k) << SYMBOL_BITS)
					     | sym | n;
			j++;
		}
	}
	return;
}

/*
 *   Creates a decode table out of gathered and adjusted frequencies, with
 * a specified table size. The decode table is simply an array where every
 * record contains three entries bitpacked.
 *
 *   Each entry has the following information:
 *
 *    +-------+--------+--------+
 *    | Range | Number | Symbol |
 *    +-------+--------+--------+
 *
 *   Implementing a frequency equivalent code.
 */

static void
build_dec_table_ac(word *freq, dword *dec_table)
{
	int i;
	int sym;
	int j = 0;

	/* Fill primary table */
	for (sym = 0; sym != MAX_SYMBOLS; sym++) {
		for (i = 0; i < freq[sym]; i++) {
			if (freq[sym])
				dec_table[j] = sym | (i << SYMBOL_BITS) |
					((freq[sym] - 1)
					 << (TABLE_BITS1 + SYMBOL_BITS));
			j++;
		}
	}
	return;
}

/*
 *   Reads frequencies and creates primary and secondary decode tables,
 * does so by reading the frequencies, adjusting them and obtaining the
 * primary and secondary decode tables. Returns the least frequent symbol
 * of the input, which will be used to identify entries to the secondary
 * decode table.
 *
 *   Once they have been created, the model for static decoding is ready.
 */

static void
build_dec_model_pic(stream *input, dword *index, dword *dec_table1,
		    dword *dec_table2, int *least_freq, dword *output_size,
		    word *hash1, word *hash2)
{	
	word freq1[MAX_SYMBOLS];
	word freq2[MAX_SYMBOLS];
	word sum1;
	word sum2;

	sum1 = read_number_pic(input, index, STATICW_SIZE - 1);
	*output_size = sum1 + 1; /* Reads total output size */
	read_freqs(input, index, freq1, freq2, &sum1, &sum2, MAX_TOT_FREQ1,
		   MAX_TOT_FREQ2);
	adjust_freqs(freq1, freq2, sum1, least_freq);

	build_dec_table_pic(freq1, dec_table1, TABLE_BITS1, hash1);
	if ((*least_freq != -1) && (freq2[(int)(*least_freq)] != 0))
		build_dec_table_pic(freq2, dec_table2, TABLE_BITS2, hash2);
	return;
}

/*
 *   Reads frequencies and creates primary and secondary decode tables,
 * building the decode model for static rf-decoding using arithmetic
 * coding in the feedback stage.
 *   Returns the least frequent symbol of the input, which will be used
 * to identify entries to the secondary decode table.
 */

static void
build_dec_model_ac(stream *input, dword *index, dword *dec_table1,
		   dword *dec_table2, int *least_freq, dword *output_size)
{	
	word freq1[MAX_SYMBOLS];
	word freq2[MAX_SYMBOLS];
	word sum1;
	word sum2;

	sum1 = read_number_pic(input, index, STATICW_SIZE - 1);
	*output_size = sum1 + 1; /* Reads total output size */
	read_freqs(input, index, freq1, freq2, &sum1, &sum2, MAX_TOT_FREQ1,
		   MAX_TOT_FREQ2);
	adjust_freqs(freq1, freq2, sum1, least_freq);

	build_dec_table_ac(freq1, dec_table1);
	if ((*least_freq != -1) && (freq2[(int)(*least_freq)] != 0))
		build_dec_table_ac(freq2, dec_table2);
	return;
}

/*
 *   Decodes a data block using the redundancy feedback algorithm with a
 * static symbol frequency distribution. Those frequency statistics are
 * retrieved from the input stream.
 *
 *   This function makes use of a primary and secondary decode table.
 * Symbols can be either common (freq1[symbol] > 0) or rare
 * (freq2[symbol] > 0). Common symbols are decoded using the primary decode
 * tables, while rare ones are decoded using the secondary decode tables.
 * This approximation is good enough to have good cache hit rates on
 * virtually any modern processor, since data tables can be kept very small,
 * fitting L1 the cache.
 *
 *   In a first instance, any codeword goes through the primary decode
 * table. In case it is identified as a low freq token (signaling the
 * encoding of a rare symbol), no output symbol is written and a second
 * codeword is pulled from the input stream. That codeword is processed
 * with the secondary decode tables and the resulting symbol is written.
 *
 *   We define the term "processing a codeword with a table" as the
 * following batch of steps:
 *
 *   1.- Retrieve the entry indexed by codeword in the table
 *   2.- Get the least significant bits of the entry, wich correspond to
 *   the symbol and output them to the output.
 *   3.- Get the next group of bits of the entry, that are the number of
 *   bits to read to complete the codeword from the input
 *   4.- Make the next codeword equal the most significant bits of the
 *   entry OR masked with the number of bits to read. This is the feedback
 *   stage
 *
 *   The bit packed structure of any decode table entry is this:
 *
 *    +-----------------------+--------------+--------+
 *    | Redundant information | Bits to read | Symbol |
 *    +-----------------------+--------------+--------+
 *
 *   The steps taken by the function are:
 *
 *   1.- Read output block size
 *   2.- Read primary and secondary frequencies
 *   3.- Adjust frequencies
 *   4.- Build decode tables
 *   5.- Until the whole output block is filled with decoded symbols:
 *
 *          5.a) - Read a codeword of CODE_BITS1 bits
 *          5.b) - If the codeword describes a rare symbol {
 *                          ** (using the primary decode tables) **
 *                    - Get the feedback bits associated to it
 *                    - Read remaining bits to make the next codeword
 *                          ** (using the secondary decode tables) **
 *                    - Get a codeword of CODE_BITS2 bits
 *                    - Output the symbol corresponding to the codeword
 *                    - Get the feedback bits associated to it
 *                    - Read remaining bits to make the next codeword
 *                 } else {
 *                          ** (using the primary decode tables) **
 *                    - Output the symbol corresponding to the codeword
 *                    - Get the feedback bits associated to it
 *                    - Read remaining bits to make the next codeword
 *                 }
 */

void
dec_static_pic(stream *input, dword *input_size, symbol *output,
	       dword *output_size, word *hash, word *identity)
{
	dword codeword;
	dword x;
	dword index = 0; /* input index */
	dword out_idx = 0;
	int rbits;
	dword end;
	dword dec_table1[CODE_RANGE1];
	dword dec_table2[TABLE_RANGE2];
	symbol sym;
	int least_freq;

	build_dec_model_pic(input, &index, dec_table1, dec_table2,
			    &least_freq, output_size, hash, identity);
	end = *output_size;
	codeword = 0;
	rbits = TABLE_BITS1;
	do { /* decoding loop */
		codeword |= read_bits(input, &index, rbits);
		codeword = dec_table1[codeword];
		sym = codeword & SYMBOL_MASK;
		codeword >>= SYMBOL_BITS;
		rbits = codeword & NBITS_MASK;
		codeword >>= NBITS_BITS;
		if (sym == least_freq) { /* Rare symbol */
			codeword |= read_bits(input, &index, rbits);
			x = codeword & (TABLE_MASK1 ^ TABLE_MASK2);
			codeword = dec_table2[codeword & TABLE_MASK2];
			sym = codeword & SYMBOL_MASK;
			codeword >>= SYMBOL_BITS;
			rbits = codeword & NBITS_MASK;
			codeword = (codeword >> NBITS_BITS) | x;
		}
		output[out_idx++] = sym;
	} while (out_idx < end);
	*input_size = index + 1;
	return;
}

/*
 *   This is an improved version of dec_static_pic(). It uses arithmetic
 * coding in the feedback stage instead of phasing in codes, trading
 * speed for compression.
 */

void
dec_static_ac(stream *input, dword *input_size, symbol *output,
	     dword *output_size)
{
	dword codeword;
	dword index = 0; /* input index */
	dword out_idx = 0;
	dword end;
	dword dec_table1[TABLE_RANGE1];
	dword dec_table2[TABLE_RANGE2];
	symbol sym;
	int least_freq;

	build_dec_model_ac(input, &index, dec_table1, dec_table2, &least_freq,
			   output_size); /* Steps 1 to 4 */
	end = *output_size;
	codeword = 0;
	do { /* decoding loop */
		codeword = read_bits(input, &index, TABLE_BITS1);
		codeword = dec_table1[codeword];
		sym = codeword & SYMBOL_MASK;
		codeword >>= SYMBOL_BITS;
		write_number_ac(input, &index, codeword & TABLE_MASK1,
			      (codeword >> TABLE_BITS1) + 1);
		if (sym == least_freq) { /* Rare symbol */
			codeword = read_bits(input, &index, TABLE_BITS2);
			codeword = dec_table2[codeword];
			sym = codeword & SYMBOL_MASK;
			codeword >>= SYMBOL_BITS;
			write_number_ac(input, &index, codeword & TABLE_MASK1,
				      (codeword >> TABLE_BITS1) + 1);
		}
		output[out_idx++] = sym;
	} while (out_idx < end);
	*input_size = index + 1;
	return;
}


