/*
 *   (c) 2003 by Eduardo Enrique Gonzlez Rodrguez
 *   Distributed under GPL, see license.txt file in the package
 *
 *   Stream encoding with dynamic frequency estimation rf-coding.
 */

#include "enc_dyn.h"
#include "freqs.h"
#include <string.h>
#include <stdlib.h>

extern int bit_mask[];
extern unsigned char bits_for_range[BR_TABLE_SIZE];

#include "out.inc"
#include "in.inc"

/*
 *   Initialization. Counts symbols to encode and fills other data structures
 *
 *   This function outputs 3 arrays containing:
 *
 *   freq: number of occurrences of each symbol
 *   
 *   address: array containing the address of each symbol, sorted by symbol
 *            value from min to max and by address value from min to max
 *            
 *   address_index: highest address associated to the symbol in the address
 *                  array
 *
 *
 *   A simple example:
 *
 *
 *   Input: 0, 0, 1, 2, 3, 2, 3, 1, 2, 3, 0, 1, 3, 2
 *
 *   Given a MAX_SYMBOLS = 6
 *
 *
 * Output:
 *
 *   freq: 3, 3, 4, 4, 0, 0.
 *   Address_index: 13, 10, 7, 3, 65535, 65535
 *   Address: 4, 6, 9, 12, 3, 5, 8, 13, 2, 7, 11, 0, 1, 10
 */

void
count_symbols(symbol *input, int input_size, word *freq, word *old_freq,
	      word *address, word *address_index)
{
	int i;
	int x;
	symbol sym;

	/* Reset frequencies */
	memset(freq, 0, MAX_SYMBOLS * sizeof(*freq));

	 /*   Count number of times each sym is seen. */
	for (i = 0; i < input_size; i++) {
		sym = input[i];
		freq[sym]++;
	}

	/*
	 *   Calculates first address of each symbol if they were put in an
	 * array, sorted from min to max.
	 */
	x = -1;
	for (i = MAX_SYMBOLS - 1; i >= 0; i--) {
		old_freq[i] = freq[i];
		address_index[i] = x;
		x += freq[i];
	}

	/*
	 *   Fills the address array with the address of all symbols, sorted
	 * by symbol value from min to max, and by address value from max to
	 * min.
	 */
	for (i = 0; i < input_size; i++) {
		sym = input[i];
		address_index[sym]++;
		address[address_index[sym]] = i;
	}
	return;
}

/*
 *   Creates a virtual data block 0 in the encoding stream. All blocks need
 * a preceding one to be processed, and this one will be used to compress
 * the first block of the stream.
 */

void
enc_dyn_block0(word *input_freq, symbol *block, int block_size, word *freq,
	       word *address, word *address_index)
{
	int i;
	unsigned char j = -1;
	word k = -1;

	scale_freq(input_freq, block_size);
	for (i = 0; i < MAX_SYMBOLS; i++) {
		address_index[i] = k;
		k += input_freq[i];
		freq[i] = input_freq[i];
	}
	for (i = 0; i < SLIDINGW_SIZE; i++) {
		do
			j++;
		while (!freq[j]);
		address[++address_index[j]] = i;
		freq[j]--;
		block[i] = j;
	}
	memcpy(freq, input_freq, sizeof(freq[0]) * MAX_SYMBOLS);
	return;
}

/* 
 *   Encodes a data block, using a dynamic estimation of the frequency
 * distribution of all symbols. This function can be used to operate on a
 * data stream by alternatively providing the ODD and EVEN values to
 * the position flag as well as swapping the freq1, address1 and
 * address_index1 arrays with freq2, address2 and address_index2.
 *
 *   Steps taken are:
 *
 *   1.- Gather symbol frequencies of the sample to be compressed
 *   2.- Adjusts offset. This is useful so this function can be used on
 *       both halves of the input buffer
 *   3.- For every symbol in the input stream:
 *           - In case it is not found in the whole sliding window:
 *                - output that symbol
 *                - output the codeword 0
 *           - Otherwise:
 *                - Determine number of msb's of the codeword that can be
 *                  phasing in encoded in the new codeword
 *                - Encode those bits in the new codeword and write the
 *                  rest to the output stream
 *
 *   4.- Restores freq2 and address_index2, modified by enc_dynamic()
 *   5.- Writes total length
 */

void
enc_dyn_pic(symbol *ibuffer1, symbol *ibuffer2, int input_size,
	    word *freq1, word *freq2,
	    word *address_index1, word *address_index2,
	    word *address1, word *address2,
	    stream *output, dword *output_idx)
{
	int i;
	dword index; /* output index */
	word old_freq[MAX_SYMBOLS];
	symbol sym;
	dword sym_freq;
	dword codeword = SLIDINGW_SIZE - 1;
	int k;

	index = *output_idx;
	output[index] = 1;
	count_symbols(ibuffer2, input_size, freq2, old_freq, address2,
		      address_index2);
	/* Adjusts freq count of first buffer */
	if (input_size > (SLIDINGW_SIZE / 2)) {
		memset(freq1, 0, MAX_SYMBOLS * sizeof(*freq1));
		for (i = input_size; i < SLIDINGW_SIZE; i++)
			freq1[(int)ibuffer1[i]]++;
	} else {
		for (i = 0; i < input_size; i++)
			freq1[(int)ibuffer1[i]]--;
	}
	for (i = input_size - 1; i >= 0; i--) {
		sym = ibuffer2[i]; /* Get sym */
		freq1[(int)ibuffer1[i]]++;
		freq2[(int)sym]--;
		address_index2[sym]--;
		sym_freq = freq1[(int)sym] + freq2[(int)sym];
		if (sym_freq == 0) {
			write_bits(output, &index, codeword, SLIDINGW_BITS);
			write_bits(output, &index, (word)sym, SYMBOL_BITS);
			codeword = 0;
		} else {
			k = FRLOG2(sym_freq);
			k += (((codeword & bit_mask[k]) | (1 << k))
			      < sym_freq)? 1: 0;
			write_bits(output, &index, codeword >> k,
				   SLIDINGW_BITS - k);
			codeword &= bit_mask[k];
			codeword = (codeword < freq2[sym])?
				address2[address_index2[sym] - codeword]:
					(address1[address_index1[sym]
					 + freq2[sym] - codeword]);
			codeword++;
		}
		codeword = (codeword + 21065) & SLIDINGW_MASK;
	}
	write_bits(output, &index, codeword, SLIDINGW_BITS);
	for (i = 0; i < MAX_SYMBOLS; i++) /* Restores address_index2 */
		address_index2[i] += old_freq[i];
	memcpy(freq2, old_freq, sizeof(old_freq)); /* And freq2 */
	write_number_pic(output, &index, input_size - 1, SLIDINGW_SIZE);
	*output_idx = index;
	return;
}

/*
 *   Modification of the previous function to feedback information using
 * arithmetic coding instead of phasing in codes.
 */

void
enc_dyn_ac(symbol *ibuffer1, symbol *ibuffer2, int input_size,
	   word *freq1, word *freq2,
	   word *address_index1, word *address_index2,
	   word *address1, word *address2,
	   stream *output, dword *output_idx)
{
	int i;
	dword index; /* output index */
	word old_freq[MAX_SYMBOLS];
	symbol sym;
	dword sym_freq;
	dword codeword = SLIDINGW_SIZE - 1;

	index = *output_idx;
	output[index] = -1L;
	count_symbols(ibuffer2, input_size, freq2, old_freq, address2,
		      address_index2);
	/* Adjusts freq count of first buffer */
	if (input_size > (SLIDINGW_SIZE / 2)) {
		memset(freq1, 0, MAX_SYMBOLS * sizeof(*freq1));
		for (i = input_size; i < SLIDINGW_SIZE; i++)
			freq1[(int)ibuffer1[i]]++;
	} else {
		for (i = 0; i < input_size; i++)
			freq1[(int)ibuffer1[i]]--;
	}
	for (i = input_size - 1; i >= 0; i--) {
		sym = ibuffer2[i]; /* Get sym */
		freq1[(int)ibuffer1[i]]++;
		freq2[(int)sym]--;
		address_index2[sym]--;
		sym_freq = freq1[(int)sym] + freq2[(int)sym];
		if (sym_freq == 0) {
			write_bits(output, &index, (word)sym, SYMBOL_BITS);
			write_bits(output, &index, 0, SLIDINGW_BITS);
		} else {
			codeword = read_number_ac(output, &index, sym_freq);
			codeword = (codeword < freq2[sym])?
				address2[address_index2[sym] - codeword]:
					(address1[address_index1[sym]
					 + freq2[sym] - codeword]);
			write_bits(output, &index, codeword + 1,
				   SLIDINGW_BITS);
		}
	}
	for (i = 0; i < MAX_SYMBOLS; i++) /* Restores address_index2 */
		address_index2[i] += old_freq[i];
	memcpy(freq2, old_freq, sizeof(old_freq)); /* And freq2 */
	write_number_pic(output, &index, input_size - 1, SLIDINGW_SIZE);
	*output_idx = index;
	return;
}

/*
 *   Modification of the previous function to work without any previous
 * block processed before. It has to be called to encode the very first
 * block of symbols of any input stream.
 */

void
enc_dyn1st(symbol *ibuffer, int input_size, word *freq,
	    word *address_index, word *address,
	    stream *output, dword *output_idx)
{
	int i;
	dword index; /* output index */
	word old_freq[MAX_SYMBOLS];
	int sym;
	dword codeword = SLIDINGW_SIZE - 1;

	index = *output_idx;
	output[index] = -1L;
	count_symbols(ibuffer, input_size, freq, old_freq, address,
		      address_index);
	for (i = input_size - 1; i > 0; i--) {
		sym = ibuffer[i]; /* Get sym */
		freq[sym]--;
		address_index[sym]--;
		if (freq[sym] == 0) {
			write_bits(output, &index, sym, SYMBOL_BITS);
			write_number_ac(output, &index, 0, i + 1);
		} else {
			codeword = read_number_ac(output, &index, freq[sym]);
			codeword = address[address_index[sym] - codeword];
			write_number_ac(output, &index, codeword + 1, i + 1);
		}
	}
	sym = ibuffer[0]; /* Get sym */
	freq[sym]--;
	address_index[sym]--;
	write_bits(output, &index, sym, SYMBOL_BITS);
	for (i = 0; i < MAX_SYMBOLS; i++) /* Restores address_index */
		address_index[i] += old_freq[i];
	memcpy(freq, old_freq, sizeof(old_freq)); /* And freq */
	write_number_pic(output, &index, input_size - 1, SLIDINGW_SIZE);
	*output_idx = index;
	return;
}


