/*
 *   (c) 2003 by Eduardo Enrique Gonzlez Rodrguez
 *   Distributed under GPL, see license.txt file in the package
 *
 *   Implementation of static frequency estimation encoding algorithms.
 */

#include "enc_stat.h"
#include "freqs.h"
#include <string.h>
#include <stdlib.h>

extern int bit_mask[];
extern unsigned char bits_for_range[BR_TABLE_SIZE];

#include "out.inc"
#include "in.inc"

/* 
 *   Gather symbol frequencies. For all the symbols of the alphabet,
 *
 *   freq1[x] <- CODE_BITS1 msb's of freq[x]
 *   freq2[x] <- CODE_BITS2 lsb's of freq[x], if freq1[x] == 0
 *
 *   After the assignment, both are normalized in order to ensure that:
 *      1) sum(freq1) = CODE_RANGE1 & sum(freq2) = CODE_RANGE2
 *      2) if (freq[x] != 0 & freq1[x] == 0) then freq2[x] > 0
 */

static void
gather_freqs(symbol *input, word input_size, word *freq1, word *freq2,
	     word *total1)
{
	int i;
	word freq[MAX_SYMBOLS];
	word sum1;
	word sum2;
	word most_freq;
	word x;
	word rem;
	int sym;

	memset(freq, 0, sizeof(freq));
	for (i = 0; i < input_size; i++) /* Count total frequency */
		freq[(int)input[i]]++;
	sum1 = 0;
	sum2 = 0;
	/*
	 *  Extracts freq1 & makes freq2[i] = 1 for those cases where freq2
	 * must be, at least, 1
	 */
	rem = MAX_SYMBOLS;
	for (i = 0; i < MAX_SYMBOLS; i++) { /* Scale freq1 */
		freq1[i] = freq[i] >> CODE_BITS2;
		freq2[i] = (freq[i] && (!freq1[i]))? 1 : 0;
		sum1 += freq1[i];
		sum2 += (freq2[i]? freq[i] & CODE_MASK2 : 0);
		rem -= freq2[i];
	}
	/*
	 *  Delivers CODE_RANGE2 frequency between all those symbols that have
	 * been assigned freq2[i] = 1 in the previous loop. Once finished, 
	 * the sum of all freq2[i] must equal CODE_RANGE2
	 */
	most_freq = 0;
	for (i = 0; i < MAX_SYMBOLS; i++) {
		if (freq2[i]) {
			x = (rem * (freq[i] & CODE_MASK2)) / sum2;
			sum2 -= (freq[i] & CODE_MASK2);
			freq2[i] += x;
			rem -= x;
			if (freq2[i] > freq2[(int)most_freq])
				most_freq = i;
		}
	}
	if (freq2[most_freq])
		freq2[most_freq] += rem;
	*total1 = sum1; /* All freq1's added together */
	return;
}

/*
 *   Create encode tables out of already gathered and adjusted frequencies.
 *
 *   Any number x with a freq f will be encoded with any of the f possible
 * codewords in the range: [c, c + f)  What codeword inside the range is to
 * be chosen solely depends on the bits to feedback
 *
 *   These tables will provide for every symbol, the corresponding codeword,
 * the minimum number of bits k required to encode the redundant range of
 * the codeword and c. Remember that phasing in codes have range = 2^k + b,
 * 0 < b < 2^k, k integer, k >= 0, c = 2^k - b
 */

static void
build_enc_tables_pic(word *freq1, word *freq2, dword *enc_table1,
		     dword *enc_table2)
{
	int sym;
	int codeword1 = 0;
	int codeword2 = 0;
	int k;
	int c;

	for (sym = 0; sym != MAX_SYMBOLS; sym++) {
		k = FRLOG2(freq1[sym]);
		c = (1 << (k + 1)) - freq1[sym];
		enc_table1[sym] = (codeword1 << (NBITS_BITS + TABLE_BITS1 + 1))
				  | (c << NBITS_BITS) | k;
		codeword1 += freq1[sym];
		k = FRLOG2(freq2[sym]);
		c = (1 << (k + 1)) - freq2[sym];
		enc_table2[sym] = (codeword2 << (NBITS_BITS + TABLE_BITS1 + 1))
				  | (c << NBITS_BITS) | k;
		codeword2 += freq2[sym];
	}
	return;
}

/*
 *   This function is a simple modification over the previous one to work
 * with rf-coding with arithmetic coding of redundant information. It
 * builds the same encode tables that return for every codeword, the
 * redundant information associated to it in the for of a (number, range)
 * pair.
 */

static void
build_enc_tables_ac(word *freq1, word *freq2, dword *enc_table1,
		    dword *enc_table2)
{
	int sym;
	int codeword1 = 0;
	int codeword2 = 0;

	for (sym = 0; sym != MAX_SYMBOLS; sym++) {
		enc_table1[sym] = (codeword1 << (TABLE_BITS1 + 1))
			| freq1[sym];
		codeword1 += freq1[sym];
		enc_table2[sym] = (codeword2 << (TABLE_BITS1 + 1))
			| freq2[sym];
		codeword2 += freq2[sym];
	}
	return;
}

/*
 *   Encodes symbol sym in codeword. Given Cin = value of codeword entering
 * the function and Cout = value of codeword exiting it, Cout encodes
 * symbol and the most significant bits of Cin. Those msb's are encoded with
 * a phasing in code. Letters used are the same as described in the
 * documentation of phasing in codes.
 *
 *   The encode table is used to supply the possible values that Cout can
 * take. table_bits is the number of valid bits of Cin.
 */

static void
enc_symbol_pic(symbol sym, dword *codeword, dword *encode_table,
	       int table_bits, stream *output, dword *index)
{
	int x;
	int c;
	int fbits;

	x = encode_table[sym];
	fbits = x & NBITS_MASK;
	x >>= NBITS_BITS;
	c = x & ((1 << (TABLE_BITS1 + 1)) - 1);
	x >>= (TABLE_BITS1 + 1);
	if ((*codeword >> (table_bits - fbits)) < c) {
		write_bits(output, index, bit_mask[table_bits - fbits]
			   & *codeword, table_bits - fbits);
		*codeword = (*codeword >> (table_bits - fbits)) + x;
	} else {
		fbits++;
		write_bits(output, index, bit_mask[table_bits - fbits]
			   & *codeword, table_bits - fbits);
		*codeword = (*codeword >> (table_bits - fbits)) - c + x;
	}
	return;
}

/*
 *   Variation of the previous function using arithmetic coding to encode
 * numbers in a range instead of phasing in codes.
 */

static void
enc_symbol_arith(symbol sym, dword *codeword, dword *encode_table,
		 stream *output, dword *index)
{
	int x;

	x = encode_table[sym];
	*codeword = read_number_ac(output, index,
				   x & ((1 << (TABLE_BITS1 + 1)) - 1))
				   + (x >> (TABLE_BITS1 + 1));
	return;
}

/*
 *   Encodes a block of data using the redundancy feedback algorithm with a
 * static symbol frequency distribution and phasing in codes.
 *
 *   The algorithm goes like this:
 *
 *   - Build the encode tables
 *   - For every symbol in the input
 *     {
 *          If (freq1[symbol] > 0) {
 *                It is encoded using the primary encode table
 *          } else {
 *                It is encoded using the secondary encode table
 *                A special token (low_freq_symbol) is encoded using
 *                     the primary encode table. That token signals the
 *                     existence of a rare symbol
 *          }
 *     }
 *   - Write the primary and secondary frequency tables
 *   - Write the output data block size
 *
 *   A symbol is encoded issuing a call to the enc_symbol() function.
 * It changes the value of codeword to a new one that encodes in its
 * redundant information as much msb's of the last value as possible.
 * Those bits which could not be phasing in coded are written to the output
 * stream.
 */

void
enc_static_pic(symbol *input, int input_size, stream *output, dword *index,
	       word *hash)
{
	int sym;
	dword codeword;
	dword x;
	int i;
	dword encode_table1[TABLE_RANGE1];
	dword encode_table2[TABLE_RANGE2];
	word freq1[MAX_SYMBOLS];
	word freq2[MAX_SYMBOLS];
	word old_freq1[MAX_SYMBOLS];
	word old_freq2[MAX_SYMBOLS];
	int least_freq;
	word sum1;

	gather_freqs(input, input_size, freq1, freq2, &sum1);
	memcpy(old_freq1, freq1, sizeof(old_freq1));
	memcpy(old_freq2, freq2, sizeof(old_freq2));
	adjust_freqs(freq1, freq2, sum1, &least_freq);
	build_enc_tables_pic(freq1, freq2, encode_table1, encode_table2);
	i = input_size - 1;
	codeword = CODE_RANGE1 - 1;
	output[*index] = 1;
	do {
		sym = input[i--];
		if (freq2[sym] != 0) { /* freq1[sym] = 0 */
			x = codeword & (TABLE_MASK1 ^ TABLE_MASK2);
			codeword &= TABLE_MASK2;
			enc_symbol_pic(sym, &codeword, encode_table2,
				       TABLE_BITS2, output, index);
			codeword |= x;
			sym = least_freq; /* low freq token */
		}
		enc_symbol_pic(sym, &codeword, encode_table1, TABLE_BITS1,
			       output, index);
		codeword = hash[codeword];
	} while (i >= 0);
	write_bits(output, index, codeword, TABLE_BITS1);
	write_freqs(output, index, old_freq1, old_freq2, MAX_TOT_FREQ1,
		    MAX_TOT_FREQ2);
	/* Write the uncompressed size */
	write_number_pic(output, index, input_size - 1, STATICW_SIZE - 1);
	return;
}

/* 
 *   This is an implementation of rf-coding using arithmetic coding of
 * numbers in a range instead of phasing in codes. It gives higher
 * compression at the cost of processing power.
 */

void
enc_static_ac(symbol *input, int input_size, stream *output, dword *index)
{
	int sym;
	dword codeword;
	int i;
	dword encode_table1[TABLE_RANGE1];
	dword encode_table2[TABLE_RANGE2];
	word freq1[MAX_SYMBOLS];
	word freq2[MAX_SYMBOLS];
	word old_freq1[MAX_SYMBOLS];
	word old_freq2[MAX_SYMBOLS];
	int least_freq;
	word sum1;

	gather_freqs(input, input_size, freq1, freq2, &sum1);
	memcpy(old_freq1, freq1, sizeof(old_freq1));
	memcpy(old_freq2, freq2, sizeof(old_freq2));
	adjust_freqs(freq1, freq2, sum1, &least_freq);
	build_enc_tables_ac(freq1, freq2, encode_table1, encode_table2);
	i = input_size - 1;
	codeword = CODE_RANGE1 - 1;
	output[*index] = (1 << (TABLE_BITS1 + 1)) - 1;
	do {
		sym = input[i--];
		if (freq2[sym] != 0) { /* freq1[sym] = 0 */
			enc_symbol_arith(sym, &codeword, encode_table2,
					 output, index);
			write_bits(output, index, codeword, TABLE_BITS2);
			sym = least_freq; /* low freq token */
		}
		enc_symbol_arith(sym, &codeword, encode_table1,
				 output, index);
		write_bits(output, index, codeword, TABLE_BITS1);
	} while (i >= 0);
	write_freqs(output, index, old_freq1, old_freq2, MAX_TOT_FREQ1,
		    MAX_TOT_FREQ2);
	/* Write the uncompressed size */
	write_number_pic(output, index, input_size - 1, STATICW_SIZE - 1);
	return;
}

