/*
 *   (c) 2003 by Eduardo Enrique Gonzlez Rodrguez
 *   Distributed under GPL, see license.txt file in the package
 *
 *  Common frequency manipulation functions.
 *
 *  This file contains the functions which are least related to the theory
 * of rf-coding, so it probably lacks of interest for most readers.
 *
 *   The functions implemented do:
 *
 *   1.- Initialization of the bits_for_range table
 *   2.- Read and write frequency distributions
 *   3.- Scale a frequency distribution
 */

#include "defs.h"
#include "freqs.h"
#include <string.h>
#include <stdlib.h>

extern unsigned char bits_for_range[BR_TABLE_SIZE];
extern int bit_mask[];

#include "in.inc"
#include "out.inc"

/*
 *   Inits the bits_for_range table. That table returns, for every range, the
 * maximum number of bits required to write a smaller number using phasing
 * in codes.
 *
 *   For a BR_TABLE_SIZE of 32768, the contents after initialization are:
 *
 *   0, 0, 1, 2, 2, 3, 3, 3, 3, 4 (8 times), 5 (16 times), 6 (32 times),
 *   7 (64 times),   ....  , 15 (16384 times)
 */

void
init_bits_for_range(void)
{
	int i;
	unsigned char max_bits;
	int num_times = 1;

	for (i = 1; (BR_TABLE_SIZE >> i) != 0; i++)
		;
	max_bits = i - 1;
	bits_for_range[0] = 0;
	for (i = 0; i < max_bits; i++) {
		memset(bits_for_range + num_times, i, num_times);
		num_times *= 2;
	}
	memset(bits_for_range + num_times, i + 1, BR_TABLE_SIZE - num_times);
	return;
}

/*
 *   Fills common and high_freq with the most common symbols and their
 * frequencies. Uses direct insertion algorithm.
 *
 *   Symbols in the arrays are kept sorted from least to most frequent.
 */

static void
get_high_freqs(word *freq, symbol *common, word *high_freq, word max_freq)
{
	int i;
	int j;
	int k;

	if (max_freq == 0)
		return;
	memset(high_freq + 1, 0, NUM_HIGH_FREQS * sizeof(*high_freq));
	high_freq[NUM_HIGH_FREQS + 1] = max_freq;
	/* Gets N highest frequencies from freq1 & freq2 */
	for (i = 0; i < MAX_SYMBOLS; i++) {
		for (j = 1; high_freq[j] < freq[i]; j++)
			; /* Find insertion place */
		j--;
		for (k = 1; k < j; k++) { /* Shift left array */
			common[k] = common[k + 1];
			high_freq[k] = high_freq[k + 1];
		}
		common[j] = i; /* Insert frequency */
		high_freq[j] = freq[i];
	}
	return;
}

/*
 *   Outputs most common symbols as well as their frequencies.
 *
 *   Outputs them in pairs (symbol, freq), where we now the freq of the
 * pair to be written must be smaller than that of the next one. Since the
 * decoding will take place reading pairs in the opposite order, the next
 * pair to be encoded will turn the prior to be read.
 *
 *   Knowing this, we have to encode the pairs so when decoding takes place,
 * we take into account that the maximum freq we can read will equal that of
 * the last pair read (unless it's the first one, case in which the maximum
 * frequency is MAX_TOT_FREQ).
 *
 *   A simple example:
 *
 *   The most common symbols of the data block are:
 *
 *   (Symbol, freq) -> (a, 120), (b, 88), (c, 70), (d, 65)
 *
 *   So they are encoded following this instruction sequence:
 *
 *   write_number(65, 70 ...);
 *   write_number('d', MAX_SYMBOLS ...);
 *   write_number(70, 88 ...);
 *   write_number('c', MAX_SYMBOLS ...);
 *   write_number(88, 120 ...);
 *   write_number('b', MAX_SYMBOLS ...);
 *   write_number(120, MAX_TOT_FREQ ...);
 *   write_number('a', MAX_SYMBOLS ...);
 *
 *   The decoding will follow this other instruction sequence:
 *
 *   read_number(x, MAX_SYMBOLS, ...);   x <- 'a'
 *   read_number(f, MAX_TOT_FREQ, ...);  f <- 120
 *   read_number(x, MAX_SYMBOLS, ...);   x <- 'b'
 *   read_number(f, 120, ...);           f <- 88
 *   read_number(x, MAX_SYMBOLS, ...);   x <- 'c'
 *   read_number(f, 88, ...);            f <- 70
 *   read_number(x, MAX_SYMBOLS, ...);   x <- 'd'
 *   read_number(f, 70, ...);            f <- 65
 */

static void
write_common_symbols(stream *output, dword *index, symbol *common,
		     word *freq, word max_freq)
{
	int i;
	word num_freq;

	num_freq = 0;
	for (i = 0; i < NUM_HIGH_FREQS; i++) { /* Contents */
		if (freq[i]) {
			write_number_pic(output, index, freq[i], freq[i + 1]);
			write_bits(output, index, common[i], SYMBOL_BITS);
			num_freq++;
		}
	}
	write_number_pic(output, index, num_freq, MAX_SYMBOLS); /* Length */
	return;
}

/*
 *   Writes the frequency tables to the input stream.
 *
 *   It does so extracting the most common symbols, encoding them with their
 * frequencies and finally encoding the rest of symbols with their
 * corresponding frequencies. By the moment, the most common symbols are
 * set to 10; this figure works well in practice.
 * 
 *   The reason behind this apparent complexity is that it encodes the
 * frequency table in a short space. Since it is usual to have a few, high
 * frequency symbols, it's much better to encode them and then encode the
 * rest of symbols knowing their maximum frequency beforehand.
 *
 *   An easier approach, encoding pairs (symbol, freq), knowing that the
 * sum of all freqs is limited by the MAX_TOT_FREQ requires more space,
 * although is simpler both in concept and practice.
 */

void
write_freqs(stream *output, dword *index, word *freq1, word *freq2,
	    word max_tot_freq1, word max_tot_freq2)
{
	int i;
	word tot1;
	word tot2;
	symbol high_symbols1[NUM_HIGH_FREQS + 2];
	symbol high_symbols2[NUM_HIGH_FREQS + 2];
	word high_freqs1[NUM_HIGH_FREQS + 2];
	word high_freqs2[NUM_HIGH_FREQS + 2];
	word max_freq1 = 0;
	word max_freq2 = 0;

	get_high_freqs(freq1, high_symbols1, high_freqs1, max_tot_freq1);
	if (max_tot_freq2)
	get_high_freqs(freq2, high_symbols2, high_freqs2, max_tot_freq2);
	for (i = NUM_HIGH_FREQS; i > 0; i--) {
		if (high_freqs1[i])
			freq1[(int)high_symbols1[i]] = 30000;
		if ((high_freqs2[i]) && max_tot_freq2)
			freq1[(int)high_symbols2[i]] = 30000;
		max_freq1 = high_freqs1[i]? high_freqs1[i]: max_freq1;
		max_freq2 = high_freqs2[i]? high_freqs2[i]: max_freq2;
	}

	if (max_tot_freq2 == 0)
		max_freq2 = 0;
	/* Encodes symbols not in high_symbols1 or high_symbols2 */
	tot1 = 0;
	tot2 = 0;
	for (i = MAX_SYMBOLS - 1; i >= 0; i--) {
		if (freq1[i] == 30000)
			continue;
		if ((freq1[i] == 0) && max_tot_freq2) { /* Also freq2 */
			tot2 += freq2[i];
			write_number_pic(output, index, freq2[i],
				     MIN(max_freq2, tot2));
		}
		tot1 += freq1[i];
		write_number_pic(output, index, freq1[i], MIN(max_freq1, tot1));
	}
	if (max_tot_freq2)
		write_common_symbols(output, index, high_symbols2 + 1,
				     high_freqs2 + 1, max_tot_freq2);
	high_freqs1[NUM_HIGH_FREQS + 1] = max_tot_freq1;
	write_common_symbols(output, index, high_symbols1 + 1,
			     high_freqs1 + 1, max_tot_freq1);
	for (i = 1; i < (NUM_HIGH_FREQS + 1); i++)
		tot1 += high_freqs1[i];
	write_number_pic(output, index, tot1, max_tot_freq1);
	return;
}

/*
 *   Read the most frequent symbols and its associated frequencies. They are
 * stored in pairs (symbol, freq), and sorted from most to least frequent.
 *  
 *   Storing first the most common symbols and their corresponding frequencies
 * makes for a more efficient coding of the frequency tables, since most of
 * the symbols in a compressible data block tend to be a short subset of the
 * whole symbol alphabet. After a group of the highest frequencies has been
 * described, the range of the remaining ones can be greatly reduced, as well
 * as the amount of bits required for their encoding.
 *
 *   For more information on the subject, take a look at the comments of
 * write_common_symbols() and related functions, in the file enc_stat.c
 */

static void
read_common_symbols(stream *input, dword *input_index, word *freq,
		    word max_freq, word *tot, word *min_high_freq)
{
	int i;
	word num;
	word sym;
	word f = 0;

	num = read_number_pic(input, input_index, MAX_SYMBOLS);
	*tot = 0;
	for (i = 0; i < num; i++) {
		sym = read_bits(input, input_index, SYMBOL_BITS);
		f = read_number_pic(input, input_index, max_freq);
		freq[sym] = f;
		*tot += f;
		max_freq = f;
	}
	*min_high_freq = f;
	return;
}

/*
 *   Reads frequency tables from the input stream.
 *
 *   It does so reading first the highest frequencies of both the primary and
 * secondary frequency tables. Once both have been loaded, the range of the
 * next frequencies to be read is greatly diminished, making it possible to
 * encode them with a lower bit count.
 *
 *   Once all the (symbol, frequency) pairs for the most common symbols are
 * loaded, the function reads the remaining ones.
 */

void
read_freqs(stream *input, dword *index, word *freq1, word *freq2,
	   word *sum1, word *sum2, word max_tot_freq1, word max_tot_freq2)
{
	int i;
	word freq_tot1;
	word freq_tot2=0;
	word max_freq1;
	word max_freq2=0;

	*sum1 = read_number_pic(input, index, max_tot_freq1);
	memset(freq1, 0, MAX_SYMBOLS * sizeof(*freq1));
	memset(freq2, 0, MAX_SYMBOLS * sizeof(*freq2));
	read_common_symbols(input, index, freq1, max_tot_freq1, &freq_tot1,
			    &max_freq1);
	if (max_tot_freq2)
		read_common_symbols(input, index, freq2, max_tot_freq2,
				    &freq_tot2, &max_freq2);
	else
		max_freq2 = 0;
	*sum2 = max_tot_freq2;
	freq_tot1 = *sum1 - freq_tot1;
	freq_tot2 = *sum2 - freq_tot2;
	for (i = 0; i < MAX_SYMBOLS; i++) {
		if (freq1[i] | (max_tot_freq2 && freq2[i]))
			continue;
		freq1[i] = read_number_pic(input, index,
					   MIN(freq_tot1, max_freq1));
		freq_tot1 -= freq1[i];
		if ((freq1[i] == 0) && max_tot_freq2) {
			freq2[i] = read_number_pic(input, index,
						   MIN(freq_tot2, max_freq2));
			freq_tot2 -= freq2[i];
		 } else
			freq2[i] = 0;
	}
	return;
}

/*
 *   Scales freq1 and freq2 to fit to their respective CODE_RANGE1 and
 * CODE_RANGE2. Also completes both frequency tables and assigns a codeword
 * to the low frequency symbols.
 *
 *   freq1 -> CODE_BITS1 higher bits of the frequency count.
 *   freq2 -> CODE_BITS2 lower bits of the frequency count, if freq1 = 0,
 *            otherwise is reset to 0.
 *   sum1  -> sum of all freq1 values
 *   sum2  -> sum of all freq2 values
 */

void
adjust_freqs(word *freq1, word *freq2, word total1, int *least_freq)
{
	int i;
	word stotal1 = 0; /* Scaled totals (sums of scaled freqs) */
	int most_freq1 = 0; /* Most frequent symbols */
	int most_freq2 = 0;

	for (i = 0; i < MAX_SYMBOLS; i++) { /* Scale freq2 */
		freq1[i] = total1? ((freq1[i] * CODE_RANGE1) / total1) : 0;
		if (freq1[i] > freq1[most_freq1])
			most_freq1 = i;
		if (freq2[i] > freq2[most_freq2])
			most_freq2 = i;
		stotal1 += freq1[i];
	}
	/* Assign remaining freq */
	freq1[most_freq1] += (CODE_RANGE1 - stotal1);
	if ((freq2[most_freq2] != 0) && (freq1[most_freq2] == 0)) {
		freq1[most_freq2] = 1;
		freq1[most_freq1]--;
	}
	if (freq2[most_freq2])
		*least_freq = most_freq2;
	else
		*least_freq = -1;
	return;
}

/* Scales frequencies in a frequency array to fit the sliding window. */
void
scale_freq(word *freq, word total)
{
	int i;
	word stotal = 0; /* Scaled totals (sums of scaled freqs) */
	word highest_freq = 0;

	for (i = 0; i < MAX_SYMBOLS; i++) { /* Scale freq2 */
		freq[i] = total? ((freq[i] * SLIDINGW_SIZE) / total) : 0;
		stotal += freq[i];
		if (freq[i] > freq[highest_freq])
			highest_freq = i;
	}
	freq[highest_freq] += (SLIDINGW_SIZE - stotal);
	return;
}

