/*
 *   (c) 2003 by Eduardo Enrique Gonzlez Rodrguez
 *   Distributed under GPL, see license.txt file in the package
 *
 *   Contains functions to read bits and numbers in a range from a bit
 * stream.
 */

/* Reads bits from stream using method 3 (LIFO without bit counter) */
static int
read_bits(stream *input, dword *index, int nbits_to_read)
{
	unsigned int output;

	if (input[*index] >> nbits_to_read) {
		output = input[*index] & bit_mask[nbits_to_read];
		input[*index] >>= nbits_to_read;
	} else {
		output = input[++(*index)] & bit_mask[nbits_to_read];
		input[*index] = (input[*index] >> nbits_to_read) |
				   (input[*index-1] << (8 * sizeof(stream)
					  - nbits_to_read));
	}
	return(output);
}

/*
 *   Reads an integer comprised in the range [0, range) from the input.
 *
 *   The integer is encoded with a phasing in code;
 *
 *   Given range = 2^k + b; b, k integers, b < 2^k
 */

static int
read_number_pic(stream *input, dword *index, int range)
{
	int number;
	unsigned char k;
	int c;

	range++;
	k = FRLOG2(range);
	c = (1 << (k + 1)) - range; /* c = 2^k - b */
	number = read_bits(input, index, k);
	if (number < c)
		return(number);
	else
		return((number << 1) - c + read_bits(input, index, 1));
}

/*
 *   Reads an integer comprised in the range [0, range) from the input.
 *
 *   The integer is encoded arithmetically. The extraction algorithm is:
 *
 *   - Read record pointed by index
 *   - If (record < range)
 *           number = record % range
 *           record = record / range
 *     Else
 *           accumulator = record and next record combined
 *           number = accumulator % range
 *           next record = accumulator / range
 *           increase index
 */

static int
read_number_ac(stream *input, dword *index, int range)
{
	dword x;
	unsigned long long int acc; /* 64 bits in linux */

	if (input[*index] >= (range)) { /* No span */
		x = input[*index];
		input[*index] = x / range;
		return(x % range);
	} else { /* Spans through two records */
		acc = input[*index];
		acc = (acc << (sizeof(stream) * 8)) | (input[*index + 1]);
		x = acc % range;
		acc /= range;
		input[*index] = acc >> (sizeof(stream) * 8);
		input[*index + 1] = acc & STREAM_MASK;
		*index += (input[*index] == 0)? 1 : 0;
		return(x);
	}
}

