/*
 *   rfcodec (2001/2003 by Eduardo Enrique Gonzlez Rodrguez)
 *
 *
 *     ********************* IMPORTANT ************************
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *       **************************************************************
 *
 *   Source for the executable file used as a demonstration of the redundancy
 * feedback coding algorithm. Both encoding and decoding are carried out in a
 * per-byte basis, splitting the data stream to process into data blocks.
 *
 *   This file makes a simple program which is a rudimentary file archiver.
 * It works with this syntax:
 *
 *    rfcodec <c<0|1|2|3>|d<0|1|2|3>> <input file> <output file>
 *
 *   The first parameter has a letter: either c (compression) or d
 * (decompression), that identifies the action to carry out. Following the
 * letter is also a number, wich specifies the compression method:
 *
 *     0 -> static frequency estimation by blocks using phasing in codes
 *     1 -> dynamic frequency estimation by blocks using phasing in codes
 *     2 -> static frequency estimation by blocks using arithmetic coding
 *     3 -> dynamic frequency estimation by blocks using arithmetic coding
 *
 *   The next 2 parameters are self explanatory.
 *
 *   Files do not have any kind of "magic number" to identify their format,
 * but raw data, so they will be accepted on input even if they do not
 * decompress well.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "defs.h"
#include "freqs.h"
#include "enc_stat.h"
#include "enc_dyn.h"
#include "dec_stat.h"
#include "dec_dyn.h"
#include "hash.h"

unsigned char bits_for_range[BR_TABLE_SIZE];
int bit_mask[] = {0, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023,
		  2047, 4095, 8191, 16383, 32767, 65535};

/*
 *   Encodes file using static method, corresponding to c0 switch.
 *
 *   Splits the file into data blocks and processes each one separately,
 * writing the result to another file. Data blocks are independent to
 * each other and share no information.
 */

void
enc_file_static_pic(FILE *input_file, FILE *output_file)
{
	symbol input[STATICW_SIZE];
	stream output[STATICW_SIZE * 4 / 3 + MAX_SYMBOLS * 2];
	int input_size;
	dword output_idx;
	word *hash;

	hash = make_hash(TABLE_RANGE1);
	init_bits_for_range();
	do {
		input_size = fread(input, sizeof(symbol), sizeof(input),
				   input_file);
		output_idx = (sizeof(output) / sizeof(stream)) - 1;
		enc_static_pic(input, input_size, output, &output_idx, hash);
		fwrite(output + output_idx, sizeof(stream),
		       (sizeof(output) / sizeof(stream)) - output_idx,
		       output_file);
	} while (!feof(input_file));
	free(hash);
	return;
}

/*
 *   Works like the previous function, using arithmetic coding instead of
 * phasing in codes.
 */

void
enc_file_static_ac(FILE *input_file, FILE *output_file)
{
	symbol input[STATICW_SIZE];
	stream output[STATICW_SIZE * 4 / 3 + MAX_SYMBOLS * 2];
	int input_size;
	dword output_idx;

	init_bits_for_range();
	do {
		input_size = fread(input, sizeof(symbol), sizeof(input),
				   input_file);
		output_idx = (sizeof(output) / sizeof(stream)) - 1;
		enc_static_ac(input, input_size, output, &output_idx);
		fwrite(output + output_idx, sizeof(stream),
		       (sizeof(output) / sizeof(stream)) - output_idx,
		       output_file);
	} while (!feof(input_file));
	return;
}

/*
 *   Decodes file using static method, corresponding to d0 switch
 *
 *   Loads from the data stream, splitting it into data blocks, processing
 * each one independently and writing the result to another file.
 */

void
dec_file_static_pic(FILE *input_file, FILE *output_file)
{
	stream input[(STATICW_SIZE / sizeof(stream)) * 2];
	symbol output[STATICW_SIZE];
	int input_size;
	dword output_size;
	dword end_ptr;
	int i;
	word *hash;
	word *identity;

	hash = make_hash(TABLE_RANGE1);
	identity = make_identity(TABLE_RANGE1);
	init_bits_for_range();
	end_ptr = 0;
	input_size = fread(input, sizeof(stream),
			   sizeof(input) / sizeof(stream), input_file);

	while (input_size) {
		/* Decodes input and writes to disk */
		end_ptr = input_size;
		dec_static_pic(input, &end_ptr, output, &output_size, hash,
			       identity);
		fwrite(output, sizeof(symbol), output_size, output_file);
		fflush(output_file);
		/*
		 * Refills input buffer. Shifts unprocessed, previously read
		 * and yet not decoded data; then proceeds to read and fill
		 * the rest of the buffer with input
		 */
		for (i = 0; i < (input_size - end_ptr); i++)
			input[i] = input[i + end_ptr];
		input_size += fread(input + input_size - end_ptr,
				    sizeof(stream), end_ptr, input_file)
				    - end_ptr;
	}
	free(hash);
	free(identity);
	return;
}

/*
 *   Works like the previous function, using arithmetic coding instead of
 * phasing in codes.
 */

void
dec_file_static_ac(FILE *input_file, FILE *output_file)
{
	stream input[(STATICW_SIZE / sizeof(stream)) * 2];
	symbol output[STATICW_SIZE];
	int input_size;
	dword output_size;
	dword end_ptr;
	int i;

	init_bits_for_range();
	end_ptr = 0;
	input_size = fread(input, sizeof(stream),
			   sizeof(input) / sizeof(stream), input_file);
	while (input_size) {
		/* Decodes input and writes to disk */
		end_ptr = input_size;
		dec_static_ac(input, &end_ptr, output, &output_size);
		fwrite(output, sizeof(symbol), output_size, output_file);
		fflush(output_file);
		/*
		 * Refills input buffer. Shifts unprocessed, previously read
		 * and yet not decoded data; then proceeds to read and fill
		 * the rest of the buffer with input
		 */
		for (i = 0; i < (input_size - end_ptr); i++)
			input[i] = input[i + end_ptr];
		input_size += fread(input + input_size - end_ptr,
				    sizeof(stream), end_ptr, input_file)
				    - end_ptr;
	}
	return;
}

/*
 *   Encodes file using dynamic method, corresponding to c1 switch.
 *
 *   Splits the input stream into data blocks, each one of them being
 * compressed depending on the information contained in the previous
 * block. In order to encode the first block of symbols, a virtual one
 * ("block 0") with the same frequency distribution is created. That
 * frequency distribution is then written to the output.
 */

void
enc_file_dyn_pic(FILE *input_file, FILE *output_file)
{
	symbol input[2 * SLIDINGW_SIZE];
	stream output[SLIDINGW_SIZE + MAX_SYMBOLS];
	word old_freq[MAX_SYMBOLS];
	word freq1[MAX_SYMBOLS];
	word address1[SLIDINGW_SIZE];
	word address_index1[MAX_SYMBOLS];
	word freq2[MAX_SYMBOLS];
	word address2[SLIDINGW_SIZE];
	word address_index2[MAX_SYMBOLS];
	int input_size;
	dword output_idx;

	init_bits_for_range();
	input_size = fread(input, sizeof(symbol), SLIDINGW_SIZE, input_file);
	output_idx = (sizeof(output) / sizeof(stream)) - 1;
	output[output_idx] = 1;
	count_symbols(input, input_size, freq1, old_freq,
		      address1, address_index1);
	enc_dyn_block0(freq1, input + SLIDINGW_SIZE, input_size, freq2,
		       address2, address_index2);
	enc_dyn_pic(input + SLIDINGW_SIZE, input, input_size, freq2, freq1,
		    address_index2, address_index1, address2, address1,
		    output, &output_idx);
	memset(freq1, 0, sizeof(freq1));
	write_freqs(output, &output_idx, freq2, freq1, SLIDINGW_SIZE + 1,
		    0);
	memcpy(freq1, old_freq, sizeof(freq1));
	fwrite(output + output_idx, sizeof(stream),
	       (sizeof(output) / sizeof(stream)) - output_idx, output_file);
	while (!feof(input_file)) {
		/* odd block symbol */
		input_size = fread(input + SLIDINGW_SIZE, sizeof(symbol),
				   SLIDINGW_SIZE, input_file);
		if (input_size <= 0)
			break;
		output_idx = (sizeof(output) / sizeof(stream)) - 1;
		enc_dyn_pic(input, input + SLIDINGW_SIZE, input_size,
			    freq1, freq2, address_index1, address_index2,
			    address1, address2, output, &output_idx);
		fwrite(output + output_idx, sizeof(stream),
		       (sizeof(output) / sizeof(stream)) - output_idx,
		       output_file);
		/* even block symbol */
		input_size = fread(input, sizeof(symbol), SLIDINGW_SIZE,
				   input_file);
		if (input_size <= 0)
			break;
		output_idx = (sizeof(output) / sizeof(stream)) - 1;
		enc_dyn_pic(input + SLIDINGW_SIZE, input, input_size,
			    freq2, freq1, address_index2, address_index1,
			    address2, address1, output, &output_idx);
		fwrite(output + output_idx, sizeof(stream),
		       (sizeof(output) / sizeof(stream)) - output_idx,
		       output_file);
	}
	return;
}

/*
 *   Decodes file using dynamic method, corresponding to d1 switch.
 *
 *   First, a frequency distribution is loaded from the input and a virtual
 * block of symbols ("block 0") is composed with it. The first block of
 * symbols is decoded using the virtual block as a reference, and
 * subsequent ones use the past decoded block as model state information.
 *
 *   Phasing in codes are used in the feedback process.
 */

void
dec_file_dyn_pic(FILE *input_file, FILE *output_file)
{
	stream input[SLIDINGW_SIZE/sizeof(stream)*3/2];
	symbol output[SLIDINGW_SIZE];
	word counter[SLIDINGW_SIZE]; /* n-th occurrence of symbol */
	word last_counter[MAX_SYMBOLS]; /* Keep track of repetitions */
	word freq[MAX_SYMBOLS]; /* Absolute frequency */
	word nullfreq[MAX_SYMBOLS]; /* to be discarded */
	dword input_size = 0;
	dword output_idx;
	dword end_ptr = 0;
	int i;
	word x;
	word nullx;

	init_bits_for_range();
	input_size = fread(input, sizeof(stream),
			   sizeof(input) / sizeof(stream), input_file);
	read_freqs(input, &end_ptr, freq, nullfreq, &x, &nullx,
		   SLIDINGW_SIZE + 1, 0);
	x = 0;
	for (i = 0; i < MAX_SYMBOLS; i++)
		x += freq[i];
	dec_dyn_block0(freq, output + 1, x, counter + 1, last_counter);
	dec_dyn_pic(input, &end_ptr, counter, last_counter, freq,
		    output, &output_idx);
	fwrite(output + 1, sizeof(symbol), output_idx, output_file);
	for (i = 0; i < (input_size - (end_ptr + 1)); i++)
		input[i] = input[i + end_ptr + 1];
	while (input_size - (end_ptr + 1)) {
		input_size += fread(0+input + input_size - end_ptr-1,
				    sizeof(stream), end_ptr+1, input_file)
				    - (end_ptr + 1);
		end_ptr = 0;
		dec_dyn_pic(input, &end_ptr, counter, last_counter,
			    freq, output, &output_idx);
		fwrite(output + 1, sizeof(symbol), output_idx, output_file);
		/*
		 *   Shifts unprocessed, previously read and yet not decoded
		 * data to the begining of the input buffer.
		 */
		for (i = 0; i < (input_size - (end_ptr + 1)); i++)
			input[i] = input[i + end_ptr + 1];
	}
	return;
}

/*
 *   Encodes a file using rf-coding with dynamic frequency estimation and
 * arithmetic coding for the feedback stage. The first block of symbols is
 * processed with enc_dyn1st(), a modification of the enc_dyn_ac().
 * Subsequent blocks are processed with enc_dyn_ac()
 */

void
enc_file_dyn_ac(FILE *input_file, FILE *output_file)
{
	symbol input[2 * SLIDINGW_SIZE];
	stream output[SLIDINGW_SIZE + MAX_SYMBOLS];
	word freq1[MAX_SYMBOLS];
	word address1[SLIDINGW_SIZE];
	word address_index1[MAX_SYMBOLS];
	word freq2[MAX_SYMBOLS];
	word address2[SLIDINGW_SIZE];
	word address_index2[MAX_SYMBOLS];
	int input_size;
	dword output_idx;

	init_bits_for_range();
	input_size = fread(input, sizeof(symbol), SLIDINGW_SIZE, input_file);
	output_idx = (sizeof(output) / sizeof(stream)) - 1;
	enc_dyn1st(input, input_size, freq1, address_index1, address1,
		   output, &output_idx);
	fwrite(output + output_idx, sizeof(stream),
	       (sizeof(output) / sizeof(stream)) - output_idx, output_file);
	while (!feof(input_file)) {
		/* odd block symbol */
		input_size = fread(input + SLIDINGW_SIZE, sizeof(symbol),
				   SLIDINGW_SIZE, input_file);
		if (input_size <= 0)
			break;
		output_idx = (sizeof(output) / sizeof(stream)) - 1;
		enc_dyn_ac(input, input + SLIDINGW_SIZE, input_size,
			   freq1, freq2, address_index1, address_index2,
			   address1, address2, output, &output_idx);
		fwrite(output + output_idx, sizeof(stream),
		       (sizeof(output) / sizeof(stream)) - output_idx,
		       output_file);
		/* even block symbol */
		input_size = fread(input, sizeof(symbol), SLIDINGW_SIZE,
				   input_file);
		if (input_size <= 0)
			break;
		output_idx = (sizeof(output) / sizeof(stream)) - 1;
		enc_dyn_ac(input + SLIDINGW_SIZE, input, input_size,
			   freq2, freq1, address_index2, address_index1,
			   address2, address1, output, &output_idx);
		fwrite(output + output_idx, sizeof(stream),
		       (sizeof(output) / sizeof(stream)) - output_idx,
		       output_file);
	}
	return;
}

/*
 *   Decodes a file using rf-coding with dynamic frequency estimation and
 * arithmetic coding for the feedback stage. The first block of symbols is
 * processed with dec_dyn1st(), a modification of the dec_dyn_ac().
 * Subsequent blocks are processed with dec_dyn_ac()
 */

void
dec_file_dyn_ac(FILE *input_file, FILE *output_file)
{
	stream input[SLIDINGW_SIZE/sizeof(stream)*3/2];
	symbol output[SLIDINGW_SIZE];
	word counter[SLIDINGW_SIZE]; /* n-th occurrence of symbol */
	word last_counter[MAX_SYMBOLS]; /* Keep track of repetitions */
	word freq[MAX_SYMBOLS]; /* Absolute frequency */
	dword input_size = 0;
	dword output_idx;
	dword end_ptr = 1;
	int i;

	init_bits_for_range();
	input_size = fread(input + 1, sizeof(stream),
			   sizeof(input) / sizeof(stream), input_file);
	memset(freq, 0, sizeof(freq));
	dec_dyn1st(input, &end_ptr, counter, last_counter, freq,
		    output, &output_idx);
	fwrite(output + 1, sizeof(symbol), output_idx, output_file);
	for (i = 0; i < (input_size - end_ptr); i++)
		input[1 + i] = input[i + end_ptr + 1];
	while (input_size +1 - (end_ptr)) {
		input_size += fread(input + input_size - end_ptr - 1,
				    sizeof(stream), end_ptr+1, input_file)
				    - (end_ptr + 1);
		end_ptr = 1;
		dec_dyn_ac(input, &end_ptr, counter, last_counter,
			   freq, output, &output_idx);
		fwrite(output + 1, sizeof(symbol), output_idx, output_file);
		/*
		 *   Shifts unprocessed, previously read and yet not decoded
		 * data to the begining of the input buffer.
		 */
		for (i = 1; i < (1 + input_size + 1 - end_ptr); i++)
			input[i] = input[i + end_ptr + 1];
	}
	return;
}


/*
 *   Presents an error/help message in case of introducing the wrong
 * parameters, otherwise opens input and output files, invokes the
 * corresponding functions and closes them, once all the processing has been
 * done.
 */

int
main(int argc, char *argv[])
{
	FILE *input_file;
	FILE *output_file;

	if (argc != 4) {
		printf("rfcodec (c) 2003 by Eduardo Enrique Gonzlez Rodrguez\n");
		printf("rfcodec comes with ABSOLUTELY NO WARRANTY;\n");
		printf("Read the license.txt file for details;\n");
		printf("Demonstration of the redundancy feedback coding algorithm\n");
		printf("Use: rfcodec <c<0|1|2|3>|d<0|1|2|3>> <input file> <output file>\n");
		printf("\nCommands: C ---> Compress, D ---> Decompress\n");
		printf("\nMethods:\n");
		printf("0 -> Static freq. estimation with phasing in codes\n");
		printf("1 -> Dynamic freq. estimation with phasing in codes\n");
		printf("2 -> Static freq. estimation with arithmetic coding\n");
		printf("3 -> Dynamic freq. estimation with arithmetic coding\n");
		exit(1);
	}
	input_file = fopen(argv[2], "rb");
	if (input_file == NULL) {
		printf("Unable to open input: %s\n", argv[2]);
		exit(1);
	}
	output_file = fopen(argv[3], "w+");
	if (output_file == NULL) {
		printf("Unable to open output: %s\n", argv[3]);
		exit(1);
	}
	if (argv[1][1] == '0') {
		if ((*argv[1] == 'c') || (*argv[1] == 'C'))
			enc_file_static_pic(input_file, output_file);
		else if ((*argv[1] == 'd') || (*argv[1] == 'D'))
			dec_file_static_pic(input_file, output_file);
		else {
			printf("Invalid option: %s\n", argv[1]);
			exit(1);
		}
	} else if (argv[1][1] == '1') {
		if ((*argv[1] == 'c') || (*argv[1] == 'C'))
			enc_file_dyn_pic(input_file, output_file);
		else if ((*argv[1] == 'd') || (*argv[1] == 'D'))
			dec_file_dyn_pic(input_file, output_file);
		else {
			printf("Invalid option: %s\n", argv[1]);
			exit(1);
		}
	} else if (argv[1][1] == '2') {
		if ((*argv[1] == 'c') || (*argv[1] == 'C'))
			enc_file_static_ac(input_file, output_file);
		else if ((*argv[1] == 'd') || (*argv[1] == 'D'))
			dec_file_static_ac(input_file, output_file);
		else {
			printf("Invalid option: %s\n", argv[1]);
			exit(1);
		}
	} else if (argv[1][1] == '3') {
		if ((*argv[1] == 'c') || (*argv[1] == 'C'))
			enc_file_dyn_ac(input_file, output_file);
		else if ((*argv[1] == 'd') || (*argv[1] == 'D'))
			dec_file_dyn_ac(input_file, output_file);
		else {
			printf("Invalid option: %s\n", argv[1]);
			exit(1);
		}
	} else {
		printf("Invalid option: %s\n", argv[1]);
		exit(1);
	}

	fclose(input_file);
	fclose(output_file);
	return(0);
}

